/*
 * livepatch_bsc1207190
 *
 * Fix for CVE-2023-0266, bsc#1207190 (fs/compat_ioctl.c part)
 *
 *  Copyright (c) 2023 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_SND)

#if !IS_ENABLED(CONFIG_COMPAT)
#error "Live patch supports only CONFIG_COMPAT=y"
#endif

#include "bsc1207190_common.h"
#include <linux/syscalls.h>

/* klp-ccp: from fs/ioctl.c */
#include <linux/syscalls.h>
#include <linux/mm.h>
#include <linux/capability.h>
#include <linux/compat.h>
#include <linux/file.h>
#include <linux/fs.h>
#include <linux/security.h>
#include <linux/export.h>
#include <linux/uaccess.h>
#include <linux/falloc.h>
#include <linux/sched/signal.h>
#include <asm/ioctls.h>

static long (*klpe_ioctl_file_clone)(struct file *dst_file, unsigned long srcfd,
			     u64 off, u64 olen, u64 destoff);

#if defined CONFIG_COMPAT && defined(CONFIG_X86_64)

static int compat_ioctl_preallocate(struct file *file, int mode,
				    struct space_resv_32 __user *argp)
{
	struct inode *inode = file_inode(file);
	struct space_resv_32 sr;

	if (copy_from_user(&sr, argp, sizeof(sr)))
		return -EFAULT;

	switch (sr.l_whence) {
	case SEEK_SET:
		break;
	case SEEK_CUR:
		sr.l_start += file->f_pos;
		break;
	case SEEK_END:
		sr.l_start += i_size_read(inode);
		break;
	default:
		return -EINVAL;
	}

	return vfs_fallocate(file, FALLOC_FL_KEEP_SIZE, sr.l_start, sr.l_len);
}
#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif

static int (*klpe_do_vfs_ioctl)(struct file *filp, unsigned int fd,
			unsigned int cmd, unsigned long arg);

COMPAT_SYSCALL_DEFINEx(3, _klpp_ioctl, unsigned int, fd, unsigned int, cmd,
		       compat_ulong_t, arg)
{
	struct fd f = fdget(fd);
	int error;

	if (!f.file)
		return -EBADF;

	/* RED-PEN how should LSM module know it's handling 32bit? */
	error = security_file_ioctl(f.file, cmd, arg);
	if (error)
		goto out;

	switch (cmd) {
	/* FICLONE takes an int argument, so don't use compat_ptr() */
	case FICLONE:
		error = (*klpe_ioctl_file_clone)(f.file, arg, 0, 0, 0);
		break;

#if defined(CONFIG_X86_64)
	case FS_IOC_RESVSP_32:
	case FS_IOC_RESVSP64_32:
		error = compat_ioctl_preallocate(f.file, 0, compat_ptr(arg));
		break;
	case FS_IOC_UNRESVSP_32:
	case FS_IOC_UNRESVSP64_32:
		error = compat_ioctl_preallocate(f.file, FALLOC_FL_PUNCH_HOLE,
				compat_ptr(arg));
		break;
	case FS_IOC_ZERO_RANGE_32:
		error = compat_ioctl_preallocate(f.file, FALLOC_FL_ZERO_RANGE,
				compat_ptr(arg));
		break;
#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif
	default:
		error = (*klpe_do_vfs_ioctl)(f.file, fd, cmd,
				     (unsigned long)compat_ptr(arg));
		if (error != -ENOIOCTLCMD)
			break;

		/*
		 * Fix CVE-2023-0266
		 *  -1 line, +4 lines
		 */
		if (f.file->f_op->compat_ioctl &&
		    f.file->f_op->compat_ioctl == READ_ONCE(klpe_snd_ctl_ioctl_compat)) {
			error = klpp_snd_ctl_ioctl_compat(f.file, cmd, arg);
		} else if (f.file->f_op->compat_ioctl)
			error = f.file->f_op->compat_ioctl(f.file, cmd, arg);
		if (error == -ENOIOCTLCMD)
			error = -ENOTTY;
		break;
	}

 out:
	fdput(f);

	return error;
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1207190.h"
#include "../kallsyms_relocs.h"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "do_vfs_ioctl", (void *)&klpe_do_vfs_ioctl },
	{ "ioctl_file_clone", (void *)&klpe_ioctl_file_clone },
};

int livepatch_bsc1207190_fs_compat_ioctl_init(void)
{
	return __klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
}

#endif /* IS_ENABLED(CONFIG_SND) */
